/*
 * Resolve lineal equation systems
 *
 * Copyright © 1998-2002, 2004-2008 Guillem Jover <guillem@hadrons.org>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <malloc.h>
#ifdef MSDOS
#include <process.h>
#endif

#include "diags.h"

/*
 * General declarations
 */

enum error_type {
	err_param = 1,
	err_din,
	err_mem,
	err_open,
	err_file,
	err_dim,
};

enum msg_id {
	msg_inequ,
	msg_ininc,
	msg_param,
	msg_paramfile,
	msg_mem,
	msg_dim,
	msg_open,
	msg_filedim,
	msg_filecoef,
};

static struct error_msg_t {
	enum error_type type;
	char *desc;
} error_msg[] = {
	[msg_inequ] = { err_din, "wrong equation number" },
	[msg_ininc] = { err_din, "wrong incognit number" },
	[msg_param] = { err_param, "wrong parameter" },
	[msg_paramfile] = { err_param, "missing file name" },
	[msg_mem] = { err_mem, "out of memory" },
	[msg_dim] = { err_dim, "too small matrix" },
	[msg_open] = { err_open, "openning file" },
	[msg_filedim] = { err_file, "wrong file format, missing dimensions" },
	[msg_filecoef] = { err_file, "wrong file format, missing coeficients" },
};

#define errmsg(msgnum)	error_msg[msgnum].type, error_msg[msgnum].desc

typedef long double real_t;
typedef int index_t;

typedef struct {
	index_t max_row;	/* allocated rows */
	index_t max_col;	/* allocated columns */
	index_t rows;		/* current valid rows */
	index_t cols;		/* current valid columns */
	index_t *order;		/* order of terms " x y z w ... " */
	real_t *x;		/* solution */
	real_t **m;		/* matrix */
} matrix_t;

static char *program = "gauss";
static char *version = "0.1.1";

static char *file = NULL;
static int b_left = 0;
static int implicit = 0;
static int determinant = 0;

/*
 * Basic matrix working functions
 */

static void
swap_real(real_t *a, real_t *b)
{
	real_t r_tmp;

	r_tmp = *a;
	*a = *b;
	*b = r_tmp;
}

static void
matrix_swap_row(matrix_t *mat, index_t r1, index_t r2)
{
	index_t i;

	for (i = 0; i <= mat->max_col; i++)
		swap_real(&mat->m[r1][i], &mat->m[r2][i]);
}

static void
matrix_swap_col(matrix_t *mat, index_t c1, index_t c2)
{
	index_t j;

	for (j = 0; j < mat->max_row; j++)
		swap_real(&mat->m[j][c1], &mat->m[j][c2]);
}

static void
swap_index(index_t *a, index_t *b)
{
	index_t tmp;

	tmp = *a;
	*a = *b;
	*b = tmp;
}

static index_t
matrix_not_zeros_in_row(matrix_t *mat, index_t row, index_t col)
{
	index_t not_zeros = 0, i;

	for (i = col; i < mat->max_col; i++)
		if (mat->m[row][i])
			not_zeros++;

	return not_zeros;
}

static index_t
matrix_not_zeros_in_col(matrix_t *mat, index_t row, index_t col)
{
	index_t not_zeros = 0, j;

	for (j = row; j < mat->max_row; j++)
		if (mat->m[j][col])
			not_zeros++;

	return not_zeros;
}

/*
 * Main matrix funcions
 */

static matrix_t *
matrix_alloc(index_t rows, index_t cols)
{
	index_t j;
	matrix_t *mat;

	mat = malloc(sizeof(matrix_t));
	if (mat == NULL)
		die(errmsg(msg_mem));

	mat->rows = mat->max_row = rows;
	mat->cols = mat->max_col = cols;

	mat->order = malloc(sizeof(index_t)*mat->max_col);
	if (mat->order == NULL)
		die(errmsg(msg_mem));

	mat->x = malloc(sizeof(real_t)*(mat->max_col));
	if (mat->x == NULL)
		die(errmsg(msg_mem));

	mat->m = malloc(sizeof(real_t *)*(mat->max_row));
	if (mat->m == NULL)
		die(errmsg(msg_mem));
	for (j = 0; j < mat->max_row; j++) {
		mat->m[j] = malloc(sizeof(real_t)*(mat->max_col + 1));
		if (mat->m[j] == NULL)
			die(errmsg(msg_mem));
	}

	return mat;
}

static void
matrix_free(matrix_t *mat)
{
	index_t j;

	free(mat->order);
	free(mat->x);
	for (j = 0; j < mat->max_row; j++)
		free(mat->m[j]);
	free(mat->m);
	free(mat);
}

static matrix_t *
matrix_fill(void)
{
	index_t i, j;
	matrix_t *mat;

	printf("introduce number of equations : ");
	if (!scanf("%i", &i))
		die(errmsg(msg_inequ));
	if (i < 2)
		die(errmsg(msg_dim));

	printf("introduce number of incognites : ");
	if (!scanf("%i", &j))
		die(errmsg(msg_ininc));
	if (j < 2)
		die(errmsg(msg_dim));

	mat = matrix_alloc(i, j);

	printf("\n");
	for (i = 0; i < mat->max_col; i++)
		mat->order[i] = i;
	for (j = 0; j < mat->max_row; j++) {
		for (i = 0; i < mat->max_col; i++) {
			printf("a(%u,%u): ", j + 1, i + 1);
			scanf("%Lg", &mat->m[j][i]);
		}
		printf("b(%u): ", j + 1);
		scanf("%Lg", &mat->m[j][i]);
		printf("\n");
	}

	return mat;
}

static matrix_t *
matrix_load(char *fname)
{
	FILE *fp;
	index_t i, j;
	matrix_t *mat;

	fp = fopen(fname, "r");
	if (fp == NULL)
		die(errmsg(msg_open));

	if (fscanf(fp, "%d %d\n", &i, &j) != 2)
		die(errmsg(msg_filedim));

	mat = matrix_alloc(i, j);

	for (i = 0; i < mat->max_col; i++)
		mat->order[i] = i;
	for (j = 0; j < mat->max_row; j++) {
		for (i = 0; i <= mat->max_col; i++)
			if (fscanf(fp, "%Lg", &mat->m[j][i]) != 1)
				die(errmsg(msg_filecoef));
		fscanf(fp, "\n");
	}
	fclose(fp);

	return mat;
}

static void
matrix_invert_b(matrix_t *mat)
{
	index_t j;

	for (j = 0; j < mat->max_row; j++)
		mat->m[j][mat->max_col] *= -1;
}

static void
matrix_print(matrix_t *mat)
{
	index_t i, j;

	printf("\n");
	for (j = 0; j < mat->rows; j++) {
		for (i = 0; i < mat->cols; i++)
			printf("a(%u,%u)=%Lg ", j + 1, mat->order[i] + 1,
			       mat->m[j][i]);
		printf("b(%u)=%Lg\n", j + 1, mat->m[j][i]);
	}
}

static void
matrix_arrange(matrix_t *mat, index_t row_ini, index_t col_ini)
{
	index_t i, j;
	index_t nz_r, nz_c;

	if (!mat->m[row_ini][col_ini]) {
		nz_r = matrix_not_zeros_in_row(mat, row_ini, col_ini);
		nz_c = matrix_not_zeros_in_col(mat, row_ini, col_ini);
		if (nz_c)  {
			for (j = row_ini + 1; j < mat->max_row; j++) {
				if (mat->m[j][col_ini]) {
					matrix_swap_row(mat, row_ini, j);
					return;
				}
			}
		} else if (nz_r) {
			for (i = col_ini + 1; i < mat->max_col; i++) {
				if (mat->m[row_ini][i]) {
					matrix_swap_col(mat, col_ini, i);
					swap_index(&mat->order[col_ini],
						  &mat->order[i]);
					return;
				}
			}
		} else { /* (!nz_r && !nz_c) */
			for (j = row_ini + 1; j < mat->max_row; j++) {
				for (i = col_ini + 1; i < mat->max_col; i++) {
					if (mat->m[j][i]) {
						matrix_swap_row(mat, row_ini, j);
						matrix_swap_col(mat, col_ini, i);
						swap_index(&mat->order[col_ini],
							  &mat->order[i]);
						return;
					}
				}
			}
		}
	}
}

static void
gauss_method(matrix_t *mat)
{
	index_t a, i, j;

	for (a = 0; a < mat->max_row - 1; a++) {
		matrix_arrange(mat, a, a);
		for (j = a + 1; j < mat->max_row; j++) {
			for (i = mat->max_col; i >= a; i--) {
				mat->m[j][i] = mat->m[j][i] * mat->m[a][a] -
					       mat->m[a][i] * mat->m[j][a];
			}
		}
	}
}

static void
resolve_deter(matrix_t *mat)
{
	index_t i, j;

	for (j = mat->rows - 1; j >= 0 ; j--) {
		/* Independent term. */
		mat->x[j] = mat->m[j][mat->cols];
		for (i = j + 1; i < mat->cols; i++) {
			mat->x[j] -= mat->m[j][i] * mat->x[i];
		}
		/* Incognit coeficient. */
		mat->x[j] /= mat->m[j][j];
	}

	printf("\n");
	for (i = 0; i < mat->rows; i++)
		printf("X(%u) = %Lg ", mat->order[i] + 1, mat->x[i]);
	printf("\n");
}

static void
resolve_indeter(matrix_t *mat)
{
	index_t i, j;

	for (j = mat->rows - 1; j >= 0; j--) {
		/* Variable separated. */
		for (i = mat->cols; i >= j; i--)
			mat->m[j][i] /= mat->m[j][j];
		/* Rest params to right. */
		for (i = mat->cols - 1; i > j; i--)
			mat->m[j][i] *= -1;
	}

	for (j = mat->rows - 2; j >= 0; j--) {
		/* Substitute vars. */
		for (i = mat->cols; i >= j + 2; i--)
			mat->m[j][i] += mat->m[j][j + 1] * mat->m[j + 1][i];
		mat->m[j][i] = 0;
	}

	printf("\n");
	for (j = 0; j < mat->rows; j++) {
		printf("X(%d) = ", mat->order[j] + 1);
		for (i = j + 2; i < mat->cols; i++) {
			if (mat->m[j][i])
				printf("%+LgX(%d) ", mat->m[j][i],
				       mat->order[i] + 1);
		}
		printf("%+Lg\n", mat->m[j][mat->cols]);
	}
}

static void
discussion(matrix_t *mat)
{
	index_t j, grade_of_freedom = 0;

	printf("\n");
	for (j = mat->max_row - 1; j >= 0 ; j--) {
		if (!matrix_not_zeros_in_row(mat, j, 0)) {
			if (mat->m[j][mat->max_col]) {
				printf(" System Incompatible.\n");
				return;
			} else
				mat->rows--;
		}
	}

	if (mat->cols != mat->rows) {
		grade_of_freedom = mat->cols - mat->rows;
		printf(" System Compatible Indeterminated with %d %s of freedom.\n",
			grade_of_freedom, grade_of_freedom > 1 ? "grades": "grade");
		resolve_indeter(mat);
	} else {
		printf(" System Compatible Determinated.\n");
		resolve_deter(mat);
	}
}

/*
 * Main program functions
 */

static void
usage(void)
{
	printf(
	"Usage: %s [-dhilv] [-ffile_name]\n"
	"\n"
	"   -f              load matrix from a file\n"
	"   -d              resolve determinants (not available yet)\n"
	"   -i              print indeterminated compatible system solutions\n"
	"                     in implicit form (not available yet)\n"
	"   -l              independent term at left side of equal\n"
	"   -h, --help      this help\n"
	"   -v, --version   program version\n\n", program);
	exit(0);
}

static void
process_args(int argc, char **argv)
{
	if (argc == 1)
		return;

	while (*(++argv)) {
		if (!strcmp(*argv, "-d"))
			determinant = 1;
		else if (!strcmp(*argv, "-l"))
			b_left = 1;
		else if (!strcmp(*argv, "-i"))
			implicit = 1;
		else if (!strcmp(*argv, "-h") || !strcmp(*argv, "-?")
			 || !strcmp(*argv, "--help"))
			usage();
		else if (!strcmp(*argv, "-v") || !strcmp(*argv, "--version")) {
			printf("%s %s\n", program, version);
			exit(0);
		} else if (!strncmp(*argv, "-f", 2)) {
			file = *argv + 2;

			if (*file == 0) {
				error(error_msg[msg_paramfile].desc);
				usage();
			}
		} else {
			error(error_msg[msg_param].desc);
			usage();
		}
	}
}

int
main(int argc, char **argv)
{
	matrix_t *matrix;

	process_args(argc, argv);

	if (file == NULL)
		matrix = matrix_fill();
	else
		matrix = matrix_load(file);

	if (b_left)
		matrix_invert_b(matrix);

	matrix_print(matrix);
	gauss_method(matrix);
	discussion(matrix);
	matrix_print(matrix);

	matrix_free(matrix);

	return 0;
}
