/*
 * Primes
 *
 * Prime number generator, to search for patterns
 *
 * Copyright © 2000-2002, 2005, 2008 Guillem Jover <guillem@hadrons.org>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA.
 *
 */

#include <sys/types.h>
#include <sys/stat.h>

#include <fcntl.h>
#include <string.h>
#include <stdbool.h>
#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>

#include "diags.h"

static char *program = "primes";
static char *version = "0.2";

typedef unsigned long int natural_t;
typedef signed long int integer_t;

struct options {
	char *bin_file;
	int action;
	unsigned int to_add;
	unsigned int idx;
	natural_t search_num;
	unsigned int base;
	char *module;
};

struct primes_file {
	int fd;
	unsigned int entries;
};

struct primes {
	unsigned int allocated;
	unsigned int idx;
	natural_t *array, *work_array;
	natural_t *sqr_array, *work_sqr_array;
};


static char *
strbase(natural_t num, unsigned int base)
{
	char base_digits[] = "0123456789abcdefghijklmnopqrstuvwxyz";
	static char basenum[256];
	char *digit = &basenum[255];

	*digit = '\0';

	if (base > sizeof(base_digits))
		return NULL;			// doing a cool SIGSEGV or not

	while (num > 0) {
		*(--digit) = base_digits[num % base];
		num /= base;
	}

	return digit;
}

static inline void
print_prime(natural_t num, unsigned int base)
{
	if (base == 10)
		printf("%lu\n", num);
	else
		printf("%s\n", strbase(num, base));
}

static void
print_primes(struct primes *primes, unsigned int base)
{
	unsigned int np = 0;

	if (base == 10)
		printf("<primes %i>\n", primes->idx);
	else
		printf("<primes %i, base %i>\n", primes->idx, base);

	while (np < primes->idx)
		print_prime(primes->array[np++], base);
}

static bool
load_text_primes(FILE *in, struct primes *primes)
{
	char buf[512], *endp;
	int local_base = 10;

	if (feof(in))
		die(1, "input data not available");

	debug("<loading text file>\n");
	fgets(buf, sizeof(buf), in);
	if (strncmp(buf, "<primes ", 8))
		return false;

	primes->allocated += strtoull(buf + 8, &endp, 10);
	primes->array = malloc(sizeof(natural_t) * primes->allocated);
	primes->sqr_array = malloc(sizeof(natural_t) * primes->allocated);

	if (!strncmp(endp, ", base ", 7))
		local_base = strtoull(endp + 7, NULL, 10);

	primes->idx = 0;

	while (!feof(in) && primes->idx < primes->allocated) {
		natural_t p;

		fgets(buf, sizeof(buf), in);
		p = strtoull(buf, NULL, local_base);
		primes->sqr_array[primes->idx] = p * p;
		primes->array[primes->idx++] = p;
	}
	debug("<loaded %i>\n", primes->allocated);

	return true;
}

/*
 * binary find algorithm
 */
static int
find_prime(struct primes *primes, natural_t num)
{
	int l = 0, r = primes->idx, m;
	integer_t cmp;

	debug("<searching prime number %lu>\n", num);
	do {
		m = (r + l) >> 1;
		debug("<l %i, m %i, r %i>\n", l, m, r);

		cmp = num - primes->array[m];
		if (cmp < 0)
			r = m - 1;
		else if (cmp > 0)
			l = m + 1;
		else return m;

	} while (l <= r);
	debug("<search positioned on %lu>\n", primes->array[l]);

	return -l;
}

static void
search_prime(struct primes *primes, natural_t num)
{
	int idx = find_prime(primes, num);

	if (idx < 0)
		printf("<not found %lu as a prime number, lower closest idx %i>\n",
		       num, -idx);
	else
		printf("<found %lu as a prime number, idx = %i>\n", num, idx);
}

static inline bool
is_prime(struct primes *primes, natural_t p)
{
	/* prime 1 not allowed (always p mod 1 = 0) & 2 not necessary */
	natural_t *ap = primes->work_array;
	natural_t *asqr = primes->work_sqr_array;

	while (*asqr++ <= p) {
		if (p % *ap++ == 0)
			return false;
	}

	return true;
}

static void
compute_primes(struct primes *primes)
{
	natural_t n;

	debug("<computing prime numbers>\n");
	switch (primes->idx) {
	case 0:
		primes->sqr_array[primes->idx] = 1;
		primes->array[primes->idx++] = 1;
	case 1:
		primes->sqr_array[primes->idx] = 4;
		primes->array[primes->idx++] = 2;
	case 2:
		primes->sqr_array[primes->idx] = 9;
		primes->array[primes->idx++] = 3;
	default:
		n = primes->array[primes->idx - 1] + 2;
	}

	primes->work_array = primes->array + 2; // skip 1 & 2;
	primes->work_sqr_array = primes->sqr_array + 2; // skip 1 & 2;

	while (primes->idx < primes->allocated) {
		if (is_prime(primes, n)) {
			primes->sqr_array[primes->idx] = n * n;
			primes->array[primes->idx++] = n;
		}

		n += 2;
	}
}

static void
module_sub_primes(struct primes *primes, unsigned int base)
{
	unsigned int np = 0;

	if (base == 10)
		printf("<primes %i>\n", primes->idx);
	else
		printf("<primes %i, base %i>\n", primes->idx, base);

	while (np++ < primes->idx - 1) {
		print_prime(primes->array[np] - primes->array[np - 1], base);
	}
}

static void
module_primes(char *module, struct primes *primes, unsigned int base)
{
	if (module && !strcmp(module, "sub")) {
		module_sub_primes(primes, base);
	} else {
		die(3, "unrecognised module");
	}
}

static bool
open_primes(char *file_name, struct primes_file *file)
{
	file->fd = open(file_name, O_RDWR|O_CREAT, S_IRUSR|S_IWUSR|S_IRGRP);
	if (file->fd < 0)
		die(2, "opening binary file");

	if (read(file->fd, &file->entries, sizeof(file->entries)) == 0)
		file->entries = 0;

	return true;
}

static bool
load_primes(struct primes_file *in, struct primes *primes)
{
	unsigned int i;

	debug("<loading prime numbers>\n");

	if (primes->array)
		return false;

	primes->allocated += in->entries;
	primes->array = malloc(sizeof(natural_t) * primes->allocated);
	primes->sqr_array = malloc(sizeof(natural_t) * primes->allocated);
	primes->idx = in->entries;

	if (!in->entries) {
		debug("<primes number file empty>\n");
		return true;
	}

	read(in->fd, primes->array, in->entries * sizeof(natural_t));

	for (i = 0; i < primes->idx; i++) {
		primes->sqr_array[i] = primes->array[i] * primes->array[i];
	}

	debug("<loaded %i>\n", in->entries);

	return true;
}

static bool
save_primes(struct primes_file *out, struct primes *primes)
{
	int diff_primes = primes->idx - out->entries;

	debug("<saving %i prime numbers>\n", diff_primes);

	lseek(out->fd, sizeof(out->entries) + out->entries * sizeof(natural_t),
	      SEEK_SET);

	if (diff_primes) {
		write(out->fd, &primes->array[out->entries],
		      diff_primes * sizeof(natural_t));
		lseek(out->fd, 0, SEEK_SET);
		write(out->fd, &primes->idx, sizeof(primes->idx));
	}

	return true;
}

static void
usage(void)
{
	printf(
	"usage: %s [OPTIONS] file\n"
	"  -a, --add num            compute (current + add) prime numbers\n"
	"  -l, --list               list prime numbers\n"
	"  -p, --print idx          print primes[idx] number\n"
	"  -s, --search num         search a prime number\n"
	"  -m, --module action      does action on every two consecutive prime numbers\n"
	"                           available actions: sub\n"
	"  -t, --text               convert text from input to a binary file\n"
	"  -b, --base b             print prime numbers in base b\n"
	"  -h, --help               this help\n"
	"  -v, --version            version\n", program);
	exit(1);
}

/*
 * Parse arguments of program
 */
static void
arguments(struct options *opt, int argc, char **argv)
{
	if (argc == 1)
		usage();

	while (*(++argv)) {
		if (!strcmp(*argv, "-h") || !strcmp(*argv, "-?") ||
		    !strcmp(*argv, "--help")) {
			usage();
		} else if (!strcmp(*argv, "-v") || !strcmp(*argv, "--version")) {
			printf("%s %s\n", program, version);
			exit(0);
		} else if (!strcmp(*argv, "-b") || !strcmp(*argv, "--base")) {
			opt->base = strtoul(*(++argv), NULL, 10);
		} else if (!strcmp(*argv, "-t") || !strcmp(*argv, "--text")) {
			opt->action = 't';
		} else if (!strcmp(*argv, "-l") || !strcmp(*argv, "--list")) {
			opt->action = 'l';
		} else if (!strcmp(*argv, "-p") || !strcmp(*argv, "--print")) {
			opt->idx = strtoul(*(++argv), NULL, 10);
			opt->action = 'p';
		} else if (!strcmp(*argv, "-a") || !strcmp(*argv, "--add")) {
			opt->to_add = strtoul(*(++argv), NULL, 10);
			opt->action = 'a';
		} else if (!strcmp(*argv, "-s") || !strcmp(*argv, "--search")) {
			opt->search_num = strtoull(*(++argv), NULL, 10);
			opt->action = 's';
		} else if (!strcmp(*argv, "-m") || !strcmp(*argv, "--module")) {
			opt->module = *(++argv);
			opt->action = 'm';
		} else if (**argv == '-') {
			usage();
		} else opt->bin_file = *argv;
	}

	if (opt->bin_file == NULL) {
		if (*argv)
			opt->bin_file = *argv;
		else
			usage();
	}
}

int
main(int argc, char **argv)
{
	struct options options;
	struct primes_file primes_file;
	struct primes primes_list;

	memset(&options, 0, sizeof(options));
	memset(&primes_list, 0, sizeof(primes_list));

	arguments(&options, argc, argv);

	if (options.base == 0)
		options.base = 10;

	if (options.action == 'a')
		primes_list.allocated = options.to_add;

	open_primes(options.bin_file, &primes_file);

	if (options.action == 't') {
		if (!load_text_primes(stdin, &primes_list))
			die(1, "<primes n[, base b]> text signature not found");
	} else
		load_primes(&primes_file, &primes_list);

	switch (options.action) {
	case 'p':
		if (options.idx < primes_list.idx)
			print_prime(primes_list.array[options.idx],
				    options.base);
		break;
	case 's':
		search_prime(&primes_list, options.search_num);
		break;
	case 'l':
		print_primes(&primes_list, options.base);
		break;
	case 'a':
		compute_primes(&primes_list);
	case 't':
		save_primes(&primes_file, &primes_list);
		break;
	case 'm':
		module_primes(options.module, &primes_list, options.base);
		break;
	}

	close(primes_file.fd);
	free(primes_list.array);
	free(primes_list.sqr_array);

	return 0;
}
